/******************************************************************************/
/* user_emp.c       Collect performance data for user emp                     */
/*                                                                            */
/* Rags Int., Inc.                                              December 1999 */
/* (c) RISC Application Growth Solutions International, Incorporated  1999    */
/*             (817) 267-2554       rags@compuserve.com                       */
/*                                                                            */
/******************************************************************************/
#include <string.h>                                       /* string functions */
#include <stdlib.h>                                                 /* stdlib */
#include <stdio.h>                                                      /* IO */
#include <malloc.h>                                                 /* malloc */
#include <cicstype.h>                              /* CICS-supplied datatypes */
#include <cics_samples.h>                  /* CICS-supplied macro definitions */
#include <cics_emp.h>                                  /* CICS EMP data types */
#include <cics_emp_tdq.h>                            /* CICS CMF record types */
#include "user_emp.h"                        /* CICS CMF User EMP definitions */

/*------------------------------*/
/* Function Protypes            */
/*------------------------------*/
void *CICS_EMP_Lookup( IN  int                              FID,
                       IN  const struct CICS_EMP_Map        Dictionary[],
                       IN  struct CICS_EMP_PerformanceClass *SystemData,
                       OUT int                              *Mask);

/*------------------------------*/
/* the ULM record               */
/*------------------------------*/
struct CICS_EMP_Record
{
    cics_ulong_t  Clk_Counter1;                            /* Clock 1 counter */
    cics_ulong_t  Clk_Time1;                                 /* Clock 1 value */
    cics_ulong_t  Clk_Counter2;                            /* Clock 2 counter */
    cics_ulong_t  Clk_Time2;                                 /* Clock 2 value */
    cics_slong_t  Counter1r;                                     /* Counter 1 */
    cics_slong_t  Counter2r;                                     /* Counter 2 */
    cics_slong_t  Counter3r;                                     /* Counter 3 */
};

/*------------------------------*/
/* the ULM save area            */
/*------------------------------*/
struct CICS_EMP_DataArea
{
    CICS_EMP_EventTime_t     Clock1;                               /* Clock 1 */
    CICS_EMP_EventTime_t     Clock2;                               /* Clock 2 */
    cics_slong_t             Counter1;                           /* Counter 1 */
    cics_slong_t             Counter2;                           /* Counter 2 */
    cics_slong_t             Counter3;                           /* Counter 3 */
    int                      Active;                   /* Clock 1 active flag */
    struct CICS_EMP_Record   Record;                            /* The record */
};

/******************************************************************************/
/* CICS_EMP_Init()       Initialization function                              */
/******************************************************************************/
CICS_EMP_Return_t CICS_EMP_Init(
                        IN    struct CICS_EMP_RegionData       *RegionData,
                        INOUT struct CICS_EMP_Functions        *Functions,
                        INOUT void                             **SaveSpace,
                        OUT   int                              *ErrorCode,
                        IN    struct CICS_EMP_PerformanceClass *SystemData,
                        IN    struct cics_eib                  *EIB,
                        IN    const struct CICS_EMP_Map        Dictionary[])
{
  void *DataSpace;                                          /* Our data space */

    /*-----------------------------------------------------*/
    /* Get some space                                      */
    /*-----------------------------------------------------*/
    DataSpace = (void *)malloc(sizeof (struct CICS_EMP_DataArea));
    memset(DataSpace, 0, sizeof (struct CICS_EMP_DataArea));

    /*-----------------------------------------------------*/
    /* Check to see if we got it                           */
    /*-----------------------------------------------------*/
    if (DataSpace == NULL)
    {
        fprintf(stderr, "[user_emp]: Init Failed %4.4s ...\n", EIB->eibtrnid);
        *ErrorCode = 1;
        *SaveSpace = NULL;
        return CICS_EMP_RETURN_FAIL;
    }

    *SaveSpace = DataSpace;
    Functions->CICS_EMP_EnterAPI        = CICS_EMP_EnterAPI;
    Functions->CICS_EMP_WriteRecord     = CICS_EMP_WriteRecord;
    Functions->CICS_EMP_ReStart         = CICS_EMP_ReStart;
#ifdef DEBUG
    fprintf(stderr, "[user_emp]: Initialized %4.4s ...\n", EIB->eibtrnid);
#endif

    return CICS_EMP_RETURN_OK;
}

/******************************************************************************/
/* CICS_EMP_EnterAPI()       ENTER API Processing                             */
/******************************************************************************/
CICS_EMP_Return_t CICS_EMP_EnterAPI(
                         IN    struct CICS_EMP_ENTER_API_Data  *EnterData,
                         INOUT void                           **SaveSpace,
                         OUT   int                             *ErrorCode)
{
  struct CICS_EMP_DataArea    *LocalData;

    LocalData = (struct CICS_EMP_DataArea *)*SaveSpace;

    /*-----------------------------------------------------*/
    /* Check initialisation OK                             */
    /*-----------------------------------------------------*/
    if (*SaveSpace == NULL)
    {
       *ErrorCode = 1;
       return CICS_EMP_RETURN_FAIL;
    }

#ifdef DEBUG
    fprintf(stderr, "[user_emp]: ID=%i EntryName=%8.8s D1=%i D2=%i\n",
                      EnterData->ID,
                      EnterData->EntryName,
                      EnterData->Data1,
                      EnterData->Data2);
#endif

    /*-----------------------------------------------------*/
    /* Switch on the EMP ID                                */
    /*-----------------------------------------------------*/
    switch (EnterData->ID)
    {
        case WRITE_EMP:                           /* EMP 1 - write the record */
            return CICS_EMP_RETURN_WRITE;
        case CLOCK_1:                            /* EMP 2 - start or stop the */
                                              /* clock depending on EnterName */
            if (strncmp (EnterData->EntryName, EMPSTART,
                         CICS_EMP_ENTRY_NAME_LEN) == 0)
            {
                LocalData->Active = 1;
                CICS_EMP_StartClock (&(LocalData->Clock1),
                                     CICS_EMP_CLOCK_TYPE_ELAPSED);
            }

            else if (strncmp (EnterData->EntryName, EMPSTOP,
                         CICS_EMP_ENTRY_NAME_LEN) == 0)
            {
                LocalData->Active = 0;
                CICS_EMP_StopClock (&(LocalData->Clock1));
            }
            else                                      /* Write some message ? */
            {
            }
            return CICS_EMP_RETURN_OK;

        case COUNTER_1:                                   /* Add Data 1 field */
            LocalData->Counter1 += EnterData->Data1;
            return CICS_EMP_RETURN_OK;

        case COUNTER_2:                                  /* Save Data 2 field */
            LocalData->Counter2 = EnterData->Data2;
            return CICS_EMP_RETURN_OK;

        case COUNTER_3:                        /* EMP 5 - Increment a counter */
            ++(LocalData->Counter3);
            return CICS_EMP_RETURN_OK;

        case DISABLE_EMP:                       /* EMP 6 - Disable monitoring */
            free(LocalData);
            *ErrorCode = 2;
            return CICS_EMP_RETURN_DISABLE;

        case CLOCK_2:                                 /* EMP 7 - time a write */
            CICS_EMP_StartClock (&(LocalData->Clock2),
                                     CICS_EMP_CLOCK_TYPE_ELAPSED);
            return CICS_EMP_RETURN_WRITE;

        case FREE_EMP:                             /* EMP 8 - disable the ULM */
            free(LocalData);
            *ErrorCode = 3;
            return CICS_EMP_RETURN_FAIL;

        default:                                        /* Any other - ignore */
            return CICS_EMP_RETURN_OK;
    }
    return CICS_EMP_RETURN_OK;
}

/******************************************************************************/
/* CICS_EMP_WriteRecord()       Write Record function                         */
/******************************************************************************/
CICS_EMP_Return_t
CICS_EMP_WriteRecord    (OUT   void                      **UserData,
                         OUT   int                       *UserDataLength,
                         IN    struct
                               CICS_EMP_PerformanceClass *SystemData,
                         IN    int                       UnSupportedFields[],
                         INOUT void                      **SaveSpace,
                         OUT   int                       *ErrorCode,
                         IN    const struct CICS_EMP_Map Dictionary[])
{
  struct CICS_EMP_DataArea *LocalData;
  cics_char_t              *RecordType;
  cics_char_t              *UserID;
  int                       i;

#ifdef DEBUG
    fprintf(stderr, "[user_emp]: WriteRecord() starting ...\n");
#endif

    RecordType = (cics_char_t *)CICS_EMP_Lookup(CICS_EMP_FID_RECORD_TYPE,
                                         Dictionary, SystemData, NULL);
    UserID     = (cics_char_t *)CICS_EMP_Lookup(CICS_EMP_FID_USER_ID,
                                         Dictionary, SystemData, NULL);
    LocalData = (struct CICS_EMP_DataArea *)*SaveSpace;

    /*-----------------------------------------------------*/
    /* Disable monitoring if we are not a ULM              */
    /* or end of task record                               */
    /*-----------------------------------------------------*/
    if ((*RecordType != CICS_EMP_REC_TYPE_ULM) &&
        (*RecordType != CICS_EMP_REC_TYPE_END))
    {
        free(LocalData);
        *ErrorCode = 4;

#ifdef DEBUG
        fprintf(stderr, "[user_emp]: WriteRecord() RETURN_DISABLE ...\n");
#endif

        return CICS_EMP_RETURN_DISABLE;
    }

    /*-----------------------------------------------------*/
    /* If we are root or the data area does not exist      */
    /* disable ULM                                         */
    /*-----------------------------------------------------*/
    else if ((strncmp(UserID, "root", 4) == 0) || (*SaveSpace == NULL))
    {
        /*-----------------------------*/
        /* Free the memory we acquired */
        /*-----------------------------*/
        if (*SaveSpace != NULL)
        {
            free(LocalData);
        }

        *ErrorCode = 5;
        *(UserData) = NULL;
        *UserDataLength = 0;

#ifdef DEBUG
        fprintf(stderr, "[user_emp]: WriteRecord() RETURN_FAIL ...\n");
#endif

        return CICS_EMP_RETURN_FAIL;
    }

    /*-----------------------------------------------------*/
    /* Verify that required fields in the PerformanceClass */
    /* record are available                                */
    /*-----------------------------------------------------*/
    i = 0;
    while(UnSupportedFields[i] != 0)
    {
        /*----------------------------------------------------*/
        /* We are only interested in the swap count field,    */
        /* so if it is not available clean up and quit.       */
        /*----------------------------------------------------*/
        if(UnSupportedFields[i] == CICS_EMP_FID_SWAP_COUNT)
        {
            /*-----------------------------*/
            /* Free the memory we acquired */
            /*-----------------------------*/
            if (*SaveSpace != NULL)
            {
                free(LocalData);
            }

            *ErrorCode = 7;
            *(UserData) = NULL;
            *UserDataLength = 0;

#ifdef DEBUG
            fprintf(stderr, "[user_emp]: WriteRecord() RETURN_FAIL ...\n");
#endif

            return CICS_EMP_RETURN_FAIL;
        }
    }

    /*----------------------------------------------------*/
    /* Set up to return our record.  Stop our first clock */
    /*----------------------------------------------------*/
    CICS_EMP_StopClock (&(LocalData->Clock1));

    LocalData->Record.Clk_Counter1 = LocalData->Clock1.Clock.Counter;
    LocalData->Record.Clk_Time1 = LocalData->Clock1.Clock.TotalTime;

    LocalData->Record.Clk_Counter2 = LocalData->Clock2.Clock.Counter;
    LocalData->Record.Clk_Time2 = LocalData->Clock2.Clock.TotalTime;

    LocalData->Record.Counter1r = LocalData->Counter1;
    LocalData->Record.Counter2r = LocalData->Counter2;
    LocalData->Record.Counter3r = LocalData->Counter3;

    /*----------------------------------------------------*/
    /* Return the record address and length               */
    /*----------------------------------------------------*/
    *(UserData) = (void *)&(LocalData->Record);
    *UserDataLength = sizeof(struct CICS_EMP_Record);

#ifdef DEBUG
    fprintf(stderr, "[user_emp]: WriteRecord() finished ...\n");
#endif

    return CICS_EMP_RETURN_OK;
}

/******************************************************************************/
/* CICS_EMP_ReStart()       Cleanup and Restart                               */
/******************************************************************************/
CICS_EMP_Return_t
CICS_EMP_ReStart        (IN    enum CICS_EMP_Finish Action,
                         INOUT void                 **SaveSpace,
                         OUT   int                  *ErrorCode)
{
  struct CICS_EMP_DataArea    *LocalData;

    LocalData = (struct CICS_EMP_DataArea *)*SaveSpace;

/* #ifdef DEBUG
 *     fprintf(stderr, "[user_emp]: ReStart(%i) starting ...\n", Action);
 * #endif
 */
    /*----------------------------------------------------*/
    /* Check if initialisation is OK                      */
    /*----------------------------------------------------*/
    if (*SaveSpace == NULL)
    {
        *ErrorCode = 6;

/* #ifdef DEBUG
 *         fprintf(stderr, "[user_emp]: ReStart(%i) RETURN_FAIL ...\n", Action);
 * #endif
 */
        return CICS_EMP_RETURN_FAIL;
    }

    /*----------------------------------------------------*/
    /* Decide what to do based on what we are asked to do */
    /*----------------------------------------------------*/
    switch (Action)
    {

        case CICS_EMP_FINISH_RESTART:
            /*------------------------------------------------*/
            /* Just written a record - restart our clocks     */
            /* and reset the counters                         */
            /*------------------------------------------------*/
            LocalData->Counter1 = 0;
            LocalData->Counter2 = 0;
            LocalData->Counter3 = 0;

            /*---------------------------------*/
            /* Restart clock 1 if it is active */
            /*---------------------------------*/
            if (LocalData->Active);
            {
                CICS_EMP_StartClock (&(LocalData->Clock1),
                                     CICS_EMP_CLOCK_TYPE_ELAPSED);
                LocalData->Active = 0;
            }

            /*--------------------*/
            /* Stop clock 2       */
            /*--------------------*/
            CICS_EMP_StopClock (&(LocalData->Clock2));
            return CICS_EMP_RETURN_OK;

        case CICS_EMP_FINISH_CLEANUP:
            /*------------------------------------------------*/
            /* Clean up - Monitoring has detected some error, */
            /* therefore release our data space               */
            /*------------------------------------------------*/
            free(LocalData);
            return CICS_EMP_RETURN_OK;
    }

    /*----------------------------------*/
    /* Unexpected code - just report it */
    /*----------------------------------*/
    *ErrorCode = Action;

#ifdef DEBUG
    fprintf(stderr, "[user_emp]: ReStart(%i) RETURN_FAIL ...\n", Action);
#endif

    return CICS_EMP_RETURN_FAIL;
}

/******************************************************************************/
/* CICS_EMP_Lookup()                                                          */
/******************************************************************************/
void *CICS_EMP_Lookup( IN  int                               FID,
                       IN  const struct CICS_EMP_Map         Dictionary[],
                       IN  struct CICS_EMP_PerformanceClass *SystemData,
                       OUT int                              *Mask)
{
    while ( ((Dictionary->Mask & CICS_EMP_FID_MASK) != FID) &&
             (Dictionary->Mask != ~0) )
    {
        Dictionary++;
    }

    if (Dictionary->Mask == ~0)
    {
        if (Mask != NULL) *Mask = 0;
        return NULL;
    }
    else
    {
        if (Mask != NULL) *Mask = Dictionary->Mask;
        return (void *)( ((cics_char_t *)SystemData) + Dictionary->Offset );
    }
    return NULL;
}

